#!/usr/bin/perl -w

# check_quota.pl - nagios plugin to check Linux filesystem quotas using repquota
#
# (c) 2011,2014 by Frederic Krueger / fkrueger-dev-checkquota@holics.at
#
# Licensed under Creative Commons Attribution-ShareAlike 3.0 Unported
# There is no warranty of any kind, explicit or implied, for anything this software does or does not do.
# 
# Updates for this piece of software could be available under the following URL:
#   http://dev.techno.holics.at/check_quota/

# TODO add monitoring for inode limits
# TODO 

### uses
use strict;
use v5.10;
use Getopt::Std;
#use Data::Dumper;             # for debugging

use lib "/usr/lib/nagios/plugins";
use lib "/usr/lib64/nagios/plugins";
use lib "/srv/nagios/libexec";
use utils qw (%ERRORS);

## in case we don't have nagios utils.pm in one of the above directories
## simply comment out the above line with "use utils" in it, and uncomment the 5 following lines
#%ERRORS = ('UNKNOWN' ,-1,
#           'OK'      , 0,
#           'WARNING' , 1,
#           'CRITICAL', 2
#          );



###
### DONT EDIT BELOW HERE
###

our ($PROG_NAME, $PROG_VERSION, $PROG_EMAIL, $PROG_URL) = ("check_quota", "0.27", 'fkrueger-dev-checkquota@holics.at', 'http://dev.techno.holics.at/check_quota/');

## not a CHANGELOG:
## v0.26	started changelog
## v0.27	added quota-grace support, in-line use of sudo, short info and several other optimizations (thanks to Pavel V.R.)
##                 added fully (with adaptations to coding-style): p1,p2,p3,p4
##		   added partially: p5+p6 (left out most of the optimizations for better debugging lateron)
##		added a few other "features" (incl. removing bugs and other inconveniences; use diff!)


our %cache = ();
our $cache_timeout = 300;

## func
sub usage
{
  my $msg = shift;
  $msg = "" if (!defined($msg));

  print "\nusage: check_quota.pl [[-u <username> [-m <mountpoint>] [-g <groupname> -m <mountpoint>]] [-Q] [-p] [-h]\n";
  print "\n";
  print "   -g     check quota for a specific group (specify 'all' for all groups)\n";
  print "   -u     check quota for a specific user (specify 'all' for all users)\n";
  print "   -Q     show information on users/groups without a set quota as well\n";
  print "   -m     check quota for a specific mountpoint (specify 'all' for all mps; which is the default)\n";
  print "   -s     short output (hide user/group 'OK' status)\n";
  print "   -p     add perfdata\n";
  print "   -h     this help\n";
  print "\n";
  print "This plugin allows you to check the linux filesystem quota for a specific user, group or all quota users.\n";
  print "\n";
  print "$PROG_NAME v$PROG_VERSION is licensed under Creative Commons Attribution-ShareAlike 3.0 Unported License.
There is no warranty of any kind, explicit or implied, for anything this software does or does not do.

The main page for this plugin can be found at: $PROG_URL

(c) 2011,2014 by Frederic Krueger / $PROG_EMAIL

";

  if ($msg ne "")
  {
    print "\n";
    print "Error: $msg\n";
    print "\n";
    print "\n";
  }
  exit $ERRORS{'UNKNOWN'};
} # end sub usage


sub convertintohumanreadable
{
  my $val = shift;
  $val = 0 if (!defined($val));
  my @k = ("b", "K", "M", "G", "T", "P");

  my $out = "";
  my $cnt = 0;
  while ($val > 1023)
  {
    $val /= 1024;
    $cnt++;
  }
  $out = sprintf ( (($cnt == 0)?"%i%s":"%.2f%s"), $val, $k[$cnt]);

  return ($out);
} # end sub convertintohumanreadable


## the following is added for later usage.
sub getinfo
{
  my $fname = shift;
  $fname = "" if (!defined($fname));

  my $rc = "";

  # 1. delete old cache files
  if (defined($cache{$fname}))
  {
    if ($cache{$fname}{'timestamp'} > time-$cache_timeout)
    {
      delete ($cache{$fname});
    }
  }

  # 2. read files not in cache
  if (! defined($cache{$fname}))
  {
    if (open (A, "<$fname"))
    {
      my $content = "";
      { local $/ = undef; $content = <A>; };
      close(A);

      if (defined ($content))
      {
        $cache{$fname} = { 'timestamp' => time, 'content' => $content };
      }
    }
    else
    {
      warn "getinfo($fname): Couldn't open fname '$fname': $!\n";
    }
  }

  # 3. return info from cache
  if (defined($cache{$fname}))
  {
    $rc = $cache{$fname}{'content'};
  }

  return ($rc);
} # end sub getinfo


sub mountpoint_exists
{
  my $mp = shift;
  my $searchie = shift;
  $mp = "" if (!defined($mp));
  $searchie = "" if (!defined($searchie));

  # remove trailing slash, if there
  $mp =~ s/\/$//g;
  # spaces are written as \040 in /proc/mounts!
  $mp =~ s/ /\\040/sg;

  my $rc = 0;
  my $mpinfo = getinfo ("/proc/mounts");
  if ((defined($mpinfo)) and ($mpinfo ne ""))
  {
    my @tmp = split /\n/, $mpinfo;
    for (my $i=0; ($i <= $#tmp) and ($rc == 0); $i++)
    {
      my $line = $tmp[$i];
      next if ($line !~ /$searchie/);

      my @t = split /\s+/, $line;

      if ($t[1] eq $mp)
      {
        $rc = 1;
      }
    } # end if mpinfofile is readable
  } # end if mountpoint-infofile exists

  return ($rc);
} # end sub mountpoint_exists


sub gracetohuman {
  my $grace = shift;
  my $weeks = int($grace / 604800);
  $grace -= $weeks * 604800;
  my $days  = int($grace / 86400);
  $grace -= $days * 86400;
  my $hours = int($grace / 3600);
  $grace -= $hours * 3600;
  my $min   = int($grace / 60);
  $grace -= $min * 60;
  return ("$weeks weeks" .(($days > 0) ? " $days days" : ""))	if ($weeks > 0);
  return ("$days days" .(($hours > 0) ? " $hours hours" : ""))	if ($days > 0);
  return ("$hours hours" .(($min > 0) ? " $min minutes" : ""))	if ($hours > 0);
  return ("$min minutes")					if ($min > 0);
  return ("$grace sec");
} # end sub gracetohuman
 




## init
use vars qw($opt_g $opt_u $opt_v $opt_h $opt_p $opt_m $opt_Q $opt_s
            $username $groupname $mountpoint $quota_status $quota_inuse $quota_softlimit $quota_hardlimit
            %ERRORS @rqout @outlist
            $verb_err $out);

$username = "";
$groupname = "";

our $bin_repquota = "/usr/sbin/repquota";
if ((!defined($bin_repquota)) or (! -e $bin_repquota))
  { $bin_repquota = `which repquota 2>/dev/null`;	chomp ($bin_repquota); }

# Turn this to 1 to see reason for parameter errors (if any)
$verb_err     = 0;


## args
if ($#ARGV < 0)
{
  &usage;
}
else
{
  getopts('ag:u:m:pQhs');
}


if (defined($opt_h))
{
  usage();
}


## and the actual arg parsing
our $addperfdata = 0;	# dont add perfdata by default
if (defined($opt_p))
{
  $addperfdata = 1;
}

our $shownoquotaalso = 0;	# only show users/groups with quotas by default
if (defined($opt_Q))
{
  $shownoquotaalso = 1;
}

our $showonlynotok = 0;		# only show users/groups with quotas, that do not have status OK
if (defined($opt_s))
{
  $showonlynotok = 1;
}

if ( (defined($opt_u)) and ($opt_u ne "") )
{
  $username = $opt_u;
  $username =~ s/[^0-9_a-zA-Z]//isg;
}
# it's either user or group, never both.
elsif ( (defined($opt_g)) and ($opt_g ne "") )
{
  $groupname = $opt_g;
  $groupname =~ s/[^0-9_a-zA-Z]//isg;
}



if ( (defined($opt_m)) and ($opt_m ne "") )
{
  if (($opt_m eq "all") or (mountpoint_exists ($opt_m, "quota")))
  {
    $mountpoint = $opt_m;
  }
  else
  {
    usage ("Supplied mountpoint '$opt_m' doesn't exist / have quota.. Exitting!");
    exit($ERRORS{'UNKNOWN'});
  }
}



###
### main
###

# get the repquota info we want to check
my $cmd = "$bin_repquota -c -v -s -p ";
if ((defined($username)) and ($username ne ""))
{
  $cmd .= "-u ";
}
elsif ((defined($groupname)) and ($groupname ne ""))
{
  $cmd .= "-g ";
}
else
{
  usage();
}

if ((defined($mountpoint)) and ($mountpoint ne "") and ($mountpoint ne "all"))
{
  $cmd .= "$mountpoint ";
}
else
{
  $cmd .= "-a ";
}

#print "executing cmd '$cmd'\n\n";

# create tempfile from sudo-call, just in case it goes wrong.. sudo returns 0 on success and 0 if not ;-)
my @rqout = split /\n/, `LC_ALL=C sudo -n -u root -- $cmd`;
my $status = $?;

if ($status != 0)
{
    usage ("Unable to run repquota binary '$bin_repquota'.\n       Are we allowed to run (via sudo) '$cmd' ?\n\nYou might want to add the following to your /etc/sudoers:\n\n" .$ENV{'USER'}. " ALL=(root) NOPASSWD: $bin_repquota *");
    exit ($ERRORS{'UNKNOWN'});
}

@rqout = sort { $a cmp $b } @rqout;


# and parse with it
my $perfdata = "";

my $nummatchingentities = 0;    # users/groups

my $last_exitcode = -1;
my $last_exitstatus = "UNKNOWN";
my $numoverquota = 0;

foreach my $line (@rqout)
{
  if (defined($opt_u))
  {
    next if (($username ne "") and ($username ne "all") and ($line !~ /$username/i));
  }
  elsif (defined($opt_g))
  {
    next if (($groupname ne "") and ($groupname ne "all") and ($line !~ /$groupname/i));
  }

  my @dat = split /\s+/, $line;
  next if (($#dat < 7) or ($dat[4] !~ /^\d+/));

  my ($quotaname, $quota_status, $quota_inuse, $quota_softlimit, $quota_hardlimit,$quota_grace) = ( $dat[0], $dat[1], $dat[2], $dat[3], $dat[4], $dat[5] );

  # remove the # in getpwnam output as 'user_#122', so rrdtool doesn't hiccup lateron.
  if ($quotaname =~ /#/)
  {
    $quotaname =~ s/#//g;
    $quotaname = "${quotaname}";
  }
  my $quotaname_perfdata = $quotaname;	$quotaname_perfdata =~ s/^uid /uid/g;

  # convert input \d+G by \d+M (but without the M)
  $quota_inuse = ($quota_inuse =~ /^(.*)[Gg]$/) ? $1*1024*1024*1024 : ($quota_inuse =~ /^(.*)[Mm]$/) ? $1*1024*1024 : ($quota_inuse =~ /^(.*)[Kk]$/) ? $1*1024 : $quota_inuse;
  $quota_softlimit = ($quota_softlimit =~ /^(.*)[Gg]$/) ? $1*1024*1024*1024 : ($quota_softlimit =~ /^(.*)[Mm]$/) ? $1*1024*1024 : ($quota_softlimit =~ /^(.*)[Kk]$/) ? $1*1024 : $quota_softlimit;
  $quota_hardlimit = ($quota_hardlimit =~ /^(.*)[Gg]$/) ? $1*1024*1024*1024 : ($quota_hardlimit =~ /^(.*)[Mm]$/) ? $1*1024*1024 : ($quota_hardlimit =~ /^(.*)[Kk]$/) ? $1*1024 : $quota_hardlimit;

  # hardlimit > 1 only shows up for users that actually have a quota (via edquota)
  if ((defined($quota_status)) and (
        	  ((defined($quota_softlimit)) and (($shownoquotaalso) or ($quota_softlimit > 0))) or
        	  ((defined($quota_hardlimit)) and (($shownoquotaalso) or ($quota_hardlimit > 0)))
      ))
  {
    # Obviously the quota_status is set by what _you_ defined via edquota for the specific user/group
    #
    # XXX Thanks to Pavel V.R. there is now a more useful quota checking at this point, that includes quota gracetime (if active)
    # XXX Also, softlimit and hardlimits are now handled in a less pointless way below here.
    # 
    # Message me if there is anything else you want and I might add it to the script. ;-)

    my $retcode = $ERRORS {'UNKNOWN'};
    $quota_status = "OK"		if ($quota_status eq "--");
    $quota_status = "WARNING"		if ($quota_status eq "-+");
    $quota_status = "WARNING"		if ($quota_status eq "+-");
    $quota_status = "CRITICAL"	if ($quota_status eq "++");
    $quota_status = "CRITICAL"	if (($quota_status eq "WARNING") and ($quota_softlimit == 0));
    if ((defined($quota_grace)) and ($quota_grace > 0))
    {
      $quota_grace -= time();
      $quota_status = "CRITICAL"	if ($quota_grace < 0);
    } # end if quota-gracetime.. is being graced.. by a gazelle, obviously.. yes, not every pun is a good pun or (at first writing) even intended ;-)

    $retcode = $ERRORS{$quota_status};

    if ((defined($quotaname)) and ($quotaname ne ""))
    {
      my $inusevalue = convertintohumanreadable ($quota_inuse);
      my $maxvalue = -1;
      my $maxvalue_hr = "";
      if ($quota_softlimit > 0)
      {	if ($quota_hardlimit > 0)
          { $maxvalue = $quota_hardlimit; }
        else
          { $maxvalue = $quota_softlimit; }
      } elsif ($quota_hardlimit > 0)
        { $maxvalue = $quota_hardlimit; }
      # else both are 0, so we dont care and leave it at ""
      else { $maxvalue = -1; }

      $maxvalue_hr = convertintohumanreadable ($maxvalue)	if ($maxvalue > 0);

      $perfdata .= ((defined($opt_u)) ? "user":"group") ."_${quotaname_perfdata}=$quota_inuse;" .(($quota_softlimit > 0) ? $quota_softlimit : ""). ";" .(($quota_hardlimit > 0) ? $quota_hardlimit : ""). ";0; ";
      push @outlist, [ $retcode, $quota_status, sprintf ( "%s - %s%s: %s%s%s",
        	      $quota_status,
        	      ((defined($opt_u)) ? (($quotaname =~ /^\d+$/) ? "uid":"login") : "group") ." ",
        	      $quotaname,
        	      $inusevalue,
        	      (($maxvalue_hr ne "") ? " / $maxvalue_hr" : ""),
        	      (($quota_grace > 0) ? " (Grace: ".gracetohuman($quota_grace).")" : "")
              ) ];
      $nummatchingentities++;
      $numoverquota++		if ($retcode != 0);
    }
  }
}



my $hadoutput = 0;
my $txtoutput = "";

foreach my $out (@outlist)
{
  $hadoutput = 1;
  if ($last_exitcode < $$out[0]) { $last_exitcode = $$out[0]; $last_exitstatus = $$out[1]; }
  if ($nummatchingentities > 1)
  {
    if ( ($showonlynotok <= 0) or (($showonlynotok > 0) and ($$out[1] ne "OK")) )
      { $txtoutput .= $$out[2] ."\n"; }
  }
#  print "got exitcode $$out[0] (status: $$out[1]) with data '$$out[2]'\n";
}

#print "txtout is now:\n$txtoutput\n-------------------------\n";

my $statusline = "";
if ($nummatchingentities > 1)
{
  if ($numoverquota > 0)
  {
    $statusline = "$last_exitstatus - $numoverquota of $nummatchingentities ".((defined($opt_u)) ? "users" : "groups")." is overquota." 
  }
  else
  {
    $statusline = "$last_exitstatus - All $nummatchingentities ".((defined($opt_u)) ? "users" : "groups")." are ok." 
  }
  $statusline .= (($addperfdata > 0) ? " | $perfdata" : "") ."\n";
} # end if many quotas checked

elsif ($nummatchingentities == 1)	# no statusline for only one result
{
  $statusline = $outlist[0][2] . (($addperfdata > 0) ? " | $perfdata" : "") ."\n";
} # end if one quota checked

else

{
  my $entfound = 1;		# provide entityfound = yes, if all users or groups were chosen
  if (($username ne "all") and ($groupname ne "all"))
  {
    $entfound = ((defined($opt_u)) ? (getpwnam ($username)) : (getgrnam ($groupname)));
    $entfound = 0 if (!defined($entfound));
  }

  if (defined($opt_u))
  {
    if ($entfound <= 0)
    {
      if ($username eq "all")
      {
        $statusline .= "UNKNOWN - No users could be found?!\n";
      }
      else
      {
        $statusline .= "UNKNOWN - No such user '$username'\n";
      }
    }
    else
    {
      if ($username eq "all")
      {
        $statusline .= "UNKNOWN - No quota for any user could be found?!\n";
      }
      else
      {
        $statusline .= "UNKNOWN - No quota for user '$username'\n";
      }
    }
  } # end if user
  elsif (defined($opt_g))
  {
    if ($entfound <= 0)
    {
      if ($groupname eq "all")
      {
        $statusline .= "UNKNOWN - No groups could be found.\n";
      }
      else
      {
        $statusline .= "UNKNOWN - No such group '$groupname'\n";
      }
    }
    else
    {
      if ($groupname eq "all")
      {
        $statusline .= "UNKNOWN - No quota for any group could be found?!\n";
      }
      else
      {
        $statusline .= "UNKNOWN - No quota for group '$groupname'\n";
      }
    }
  } # end if group
} # end if no quotas checked

print "$statusline$txtoutput";
exit $last_exitcode;


